///
///  TraceLogTests.swift
///
///  Copyright 2015 Tony Stone
///
///  Licensed under the Apache License, Version 2.0 (the "License");
///  you may not use this file except in compliance with the License.
///  You may obtain a copy of the License at
///
///  http://www.apache.org/licenses/LICENSE-2.0
///
///  Unless required by applicable law or agreed to in writing, software
///  distributed under the License is distributed on an "AS IS" BASIS,
///  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
///  See the License for the specific language governing permissions and
///  limitations under the License.
///
///  Created by Tony Stone on 11/1/15.
////
import XCTest
import Dispatch

import TraceLog

///
/// Main test class for Swift
///
class TraceLogTestsSwift: XCTestCase {

    let testTag = "Test Tag"

    // MARK: - Configuration

    func testConfigureWithNoArgs() {
        TraceLog.configure()
    }

    func testConfigureWithLogWriters() {
        let testMessage = "TraceLog Configured using: {\n\tglobal: {\n\n\t\tALL = INFO\n\t}\n}"

        let testWriter = ValidateExpectedValuesTestWriter(level: .info, tag: "TraceLog", message: testMessage)

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "INFO"])

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testConfigureWithLogWritersAndEnvironment() {
        let testMessage = "TraceLog Configured using: {\n\ttags: {\n\n\t\tTraceLog = TRACE4\n\t}\n\tprefixes: {\n\n\t\tNS = ERROR\n\t}\n\tglobal: {\n\n\t\tALL = TRACE4\n\t}\n}"

        let testWriter = ValidateExpectedValuesTestWriter(level: .info, tag: "TraceLog", message: testMessage)

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "TRACE4",
                                                                    "LOG_PREFIX_NS": "ERROR",
                                                                    "LOG_TAG_TraceLog": "TRACE4"])

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testConfigureWithLogWritersAndEnvironmentGlobalInvalidLogLevel() {
        let testEntries: [ValidateExpectedValuesTestWriter.ExpectedLogEntry] = [
            (timestamp: nil, level: .info,    tag: "TraceLog", message: "TraceLog Configured using: {\n\tglobal: {\n\n\t\tALL = INFO\n\t}\n}", runtimeContext: nil, staticContext: nil),
            (timestamp: nil, level: .warning, tag: "TraceLog", message: "Variable \'LOG_ALL\' has an invalid logLevel of \'TRACE5\'. \'LOG_ALL\' will be set to INFO.", runtimeContext: nil, staticContext: nil)
        ]

        let testWriter = ValidateExpectedValuesTestWriter(expected: testEntries)

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "TRACE5"])

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testConfigureWithLogWritersAndEnvironmentPrefixInvalidLogLevel() {
        let testEntries: [ValidateExpectedValuesTestWriter.ExpectedLogEntry] = [
            (timestamp: nil, level: .info,    tag: "TraceLog", message: "TraceLog Configured using: {\n\tglobal: {\n\n\t\tALL = INFO\n\t}\n}", runtimeContext: nil, staticContext: nil),
            (timestamp: nil, level: .warning, tag: "TraceLog", message: "Variable \'LOG_PREFIX_NS\' has an invalid logLevel of \'TRACE5\'. \'LOG_PREFIX_NS\' will NOT be set.", runtimeContext: nil, staticContext: nil)
        ]

        let testWriter = ValidateExpectedValuesTestWriter(expected: testEntries)

        TraceLog.configure(writers: [testWriter], environment: ["LOG_PREFIX_NS": "TRACE5"])

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testConfigureWithLogWritersAndEnvironmentTagInvalidLogLevel() {
        let testEntries: [ValidateExpectedValuesTestWriter.ExpectedLogEntry] = [
            (timestamp: nil, level: .info,    tag: "TraceLog", message: "TraceLog Configured using: {\n\tglobal: {\n\n\t\tALL = INFO\n\t}\n}", runtimeContext: nil, staticContext: nil),
            (timestamp: nil, level: .warning, tag: "TraceLog", message: "Variable \'LOG_TAG_TRACELOG\' has an invalid logLevel of \'TRACE5\'. \'LOG_TAG_TRACELOG\' will NOT be set.", runtimeContext: nil, staticContext: nil)
        ]

        let testWriter = ValidateExpectedValuesTestWriter(expected: testEntries)

        TraceLog.configure(writers: [testWriter], environment: ["LOG_TAG_TraceLog": "TRACE5"])

         self.wait(for: [testWriter.expectation], timeout: 2)
    }

    // MARK: ConcurrencyMode tests

    func testModeDirectIsSameThread() {
        let input: (thread: Thread, message: String) = (Thread.current, "Random test message.")

        let validatingWriter = CallbackTestWriter() { (timestamp, level, tag, message, runtimeContext, staticContext) -> Void in

            // We ignore all but the message that is equal to our input message to avoid TraceLogs startup messages
            if message == input.message {
                XCTAssertEqual(Thread.current, input.thread)
            }
        }

        TraceLog.configure(writers: [.direct(validatingWriter)], environment: ["LOG_ALL": "INFO"])

        logInfo { input.message }
    }

    // Note: it does not make sense to test Sync for same thread or different as there is no guarantee it will be either.

    func testModeAsyncIsDifferentThread() {
        let semaphore = DispatchSemaphore(value: 0)
        let input: (thread: Thread, message: String) = (Thread.current, "Random test message.")

        let validatingWriter = CallbackTestWriter() { (timestamp, level, tag, message, runtimeContext, staticContext) -> Void in

            // We ignore all but the message that is equal to our input message to avoid TraceLogs startup messages
            if message == input.message {
                XCTAssertNotEqual(Thread.current, input.thread)

                semaphore.signal()
            }
        }

        /// Setup test with Writer
        TraceLog.configure(writers: [.async(validatingWriter)], environment: ["LOG_ALL": "INFO"])

        /// Run test.
        logInfo { input.message }

        /// Wait for the thread to return
        XCTAssertEqual(semaphore.wait(timeout: .now() + 0.1), .success)
    }

    func testModeSyncBlocks() {
        let input: (thread: Thread, message: String) = (Thread.current, "Random test message.")
        var logged: Bool = false

        let validatingWriter = CallbackTestWriter() { (timestamp, level, tag, message, runtimeContext, staticContext) -> Void in

            // We ignore all but the message that is equal to our input message to avoid TraceLogs startup messages
            if message == input.message {
                logged = true
            }
        }

        TraceLog.configure(writers: [.sync(validatingWriter)], environment: ["LOG_ALL": "INFO"])

        /// This should block until our writer is called.
        logInfo { input.message }

        XCTAssertEqual(logged, true) /// Not a definitive test.
    }

    func testNoDeadLockDirectMode() {
        TraceLog.configure(writers: [.direct(SleepyTestWriter(sleepTime: 100))], environment: ["LOG_ALL": "INFO"])

        self._testNoDeadLock()
    }

    func testNoDeadLockSyncMode() {
        TraceLog.configure(writers: [.sync(SleepyTestWriter(sleepTime: 100))], environment: ["LOG_ALL": "INFO"])

        self._testNoDeadLock()
    }

    func testNoDeadLockAsyncMode() {
        TraceLog.configure(writers: [.async(SleepyTestWriter(sleepTime: 100))], environment: ["LOG_ALL": "INFO"])

        self._testNoDeadLock()
    }

    func _testNoDeadLock() {

        let queue   = DispatchQueue(label: "_testNoDeadLock.queue", attributes: .concurrent)
        let loggers = DispatchGroup()

        for _ in 0...20 {
            queue.async(group: loggers) {

                for _ in 0...1000 {
                    logInfo { "Random test message." }
                }
            }
        }
        /// Note: Increased this wait time to 120 for iPhone 6s iOS 9.3 which was taking a little longer to run threw the test.
        XCTAssertEqual(loggers.wait(timeout: .now() + 120.0), .success)
    }

    // MARK: - Logging Methods

    func testLogError() {
        let testMessage = "Swift: \(#function)"

        let testWriter = ValidateExpectedValuesTestWriter(level: .error, tag: self.testTag, message: testMessage)

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "ERROR"])

        logError(testTag) { testMessage }

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testLogWarning() {
        let testMessage = "Swift: \(#function)"

        let testWriter = ValidateExpectedValuesTestWriter(level: .warning, tag: self.testTag, message: testMessage)

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "WARNING"])

        logWarning(testTag) { testMessage }

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testLogInfo() {
        let testMessage = "Swift: " + #function

        let testWriter = ValidateExpectedValuesTestWriter(level: .info, tag: testTag, message: testMessage, filterTags: ["TraceLog"])

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "INFO"])

        logInfo(testTag) { "Swift: \(#function)" }

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testLogTrace() {
        let testMessage = "Swift: " + #function

        let testWriter = ValidateExpectedValuesTestWriter(level: .trace1, tag: testTag, message: testMessage, filterTags: ["TraceLog"])

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "TRACE1"])

        logTrace(testTag) { testMessage }

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testLogTrace1() {
        let testMessage = "Swift: " + #function

        let testWriter = ValidateExpectedValuesTestWriter(level: .trace1, tag: testTag, message: testMessage, filterTags: ["TraceLog"])

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "TRACE1"])

        logTrace(testTag, level: 1) { testMessage }

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testLogTrace2() {
        let testMessage = "Swift: " + #function

        let testWriter = ValidateExpectedValuesTestWriter(level: .trace2, tag: testTag, message: testMessage, filterTags: ["TraceLog"])

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "TRACE2"])

        logTrace(testTag, level: 2) { testMessage }

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testLogTrace3() {
        let testMessage = "Swift: " + #function

        let testWriter = ValidateExpectedValuesTestWriter(level: .trace3, tag: testTag, message: testMessage, filterTags: ["TraceLog"])

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "TRACE3"])

        logTrace(testTag, level: 3) { testMessage }

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    func testLogTrace4() {
        let testMessage = "Swift: " + #function

        let testWriter = ValidateExpectedValuesTestWriter(level: .trace4, tag: testTag, message: testMessage, filterTags: ["TraceLog"])

        TraceLog.configure(writers: [testWriter], environment: ["LOG_ALL": "TRACE4"])

        logTrace(testTag, level: 4) { testMessage }

        self.wait(for: [testWriter.expectation], timeout: 2)
    }

    // MARK: - Logging Methods when level below log level.

    func testLogErrorWhileOff() {
        let testMessage = "Swift: " + #function
        let semaphore = DispatchSemaphore(value: 0)

        let writer = FailWhenFiredWriter(semaphore: semaphore)

        TraceLog.configure(writers: [writer], environment: ["LOG_ALL": "OFF"])

        logError(testTag) { testMessage }

        let result = semaphore.wait(wallTimeout: DispatchWallTime.now() + .seconds(1))

        /// Note: success in this case means the test failed because the semaphore was signaled by the call to log the message.
        if result == .success {
            XCTAssertNil("Log level was OFF but message was written anyway.")
        }
    }

    func testLogWarningWhileOff() {
        let testMessage = "Swift: " + #function
        let semaphore = DispatchSemaphore(value: 0)

        let writer = FailWhenFiredWriter(semaphore: semaphore)

        TraceLog.configure(writers: [writer], environment: ["LOG_ALL": "OFF"])

        logWarning(testTag) { testMessage }

        let result = semaphore.wait(wallTimeout: DispatchWallTime.now() + .seconds(1))

        /// Note: success in this case means the test failed because the semaphore was signaled by the call to log the message.
        if result == .success {
            XCTAssertNil("Log level was OFF but message was written anyway.")
        }
    }

    func testLogInfoWhileOff() {
        let testMessage = "Swift: " + #function
        let semaphore = DispatchSemaphore(value: 0)

        let writer = FailWhenFiredWriter(semaphore: semaphore)

        TraceLog.configure(writers: [writer], environment: ["LOG_ALL": "OFF"])

        logInfo(testTag) { testMessage }

        let result = semaphore.wait(wallTimeout: DispatchWallTime.now() + .seconds(1))

        /// Note: success in this case means the test failed because the semaphore was signaled by the call to log the message.
        if result == .success {
            XCTAssertNil("Log level was OFF but message was written anyway.")
        }
    }

    func testLogTraceWhileOff() {
        let testMessage = "Swift: " + #function
        let semaphore = DispatchSemaphore(value: 0)

        let writer = FailWhenFiredWriter(semaphore: semaphore)

        TraceLog.configure(writers: [writer], environment: ["LOG_ALL": "OFF"])

        logTrace(testTag) { testMessage }

        let result = semaphore.wait(wallTimeout: DispatchWallTime.now() + .seconds(1))

        /// Note: success in this case means the test failed because the semaphore was signaled by the call to log the message.
        if result == .success {
            XCTAssertNil("Log level was OFF but message was written anyway.")
        }
    }

    func testLogTrace1WhileOff() {
        let testMessage = "Swift: " + #function
        let semaphore = DispatchSemaphore(value: 0)

        let writer = FailWhenFiredWriter(semaphore: semaphore)

        TraceLog.configure(writers: [writer], environment: ["LOG_ALL": "OFF"])

        logTrace(1, testTag) { testMessage }

        let result = semaphore.wait(wallTimeout: DispatchWallTime.now() + .seconds(1))

        /// Note: success in this case means the test failed because the semaphore was signaled by the call to log the message.
        if result == .success {
            XCTAssertNil("Log level was OFF but message was written anyway.")
        }
    }

    func testLogTrace2WhileOff() {
        let testMessage = "Swift: " + #function
        let semaphore = DispatchSemaphore(value: 0)

        let writer = FailWhenFiredWriter(semaphore: semaphore)

        TraceLog.configure(writers: [writer], environment: ["LOG_ALL": "OFF"])

        logTrace(2, testTag) { testMessage }

        let result = semaphore.wait(wallTimeout: DispatchWallTime.now() + .seconds(1))

        /// Note: success in this case means the test failed because the semaphore was signaled by the call to log the message.
        if result == .success {
            XCTAssertNil("Log level was OFF but message was written anyway.")
        }
    }

    func testLogTrace3WhileOff() {
        let testMessage = "Swift: " + #function
        let semaphore = DispatchSemaphore(value: 0)

        let writer = FailWhenFiredWriter(semaphore: semaphore)

        TraceLog.configure(writers: [writer], environment: ["LOG_ALL": "OFF"])

        logTrace(3, testTag) { testMessage }

        let result = semaphore.wait(wallTimeout: DispatchWallTime.now() + .seconds(1))

        /// Note: success in this case means the test failed because the semaphore was signaled by the call to log the message.
        if result == .success {
            XCTAssertNil("Log level was OFF but message was written anyway.")
        }
    }

    func testLogTrace4WhileOff() {
        let testMessage = "Swift: " + #function
        let semaphore = DispatchSemaphore(value: 0)

        let writer = FailWhenFiredWriter(semaphore: semaphore)

        TraceLog.configure(writers: [writer], environment: ["LOG_ALL": "OFF"])

        logTrace(4, testTag) { testMessage }

        let result = semaphore.wait(wallTimeout: DispatchWallTime.now() + .seconds(1))

        /// Note: success in this case means the test failed because the semaphore was signaled by the call to log the message.
        if result == .success {
            XCTAssertNil("Log level was OFF but message was written anyway.")
        }
    }
}
